from .. import gv as gvv

class InputValidator:

    @staticmethod
    def validate_input(df,QI_attr,Sensitive_attr,cat_indecies,verbose,max_iter,anonimize_ratio,max_cluster_distance,nan_replacement_int=0,nan_replacement_str=''):
        QI_LEN,QI,_df,SA,IS_CAT,QI_RANGE_VAL,QI_RANGE_VAL,CAT_UNIQUE,NUM_COL,CAT_COL,_DEBUG,RANGE_FIX,CAT_INDEXES,NUM_COL_RANGE,CAT_COL_RANGE = validator(
            df,QI_attr,Sensitive_attr,cat_indecies,nan_replacement_int,nan_replacement_str)

        gv = [QI, SA, IS_CAT, QI_LEN, QI_RANGE_VAL, _df,NUM_COL,NUM_COL_RANGE,CAT_COL,CAT_COL_RANGE,CAT_INDEXES,RANGE_FIX]
        gvname = ['QI', 'SA', 'IS_CAT', 'QI_LEN', 'QI_RANGE_VAL', '_df','NUM_COL','NUM_COL_RANGE','CAT_COL','CAT_COL_RANGE','CAT_INDEXES','RANGE_FIX']
        gv_dict = {}
        for i in zip(gvname,gv):
            gv_dict[i[0]] = i[1]

        gvv.init(gv_dict)

    @staticmethod
    def L_Diverse_Validate(dataframe , quasi_identifiers , sensitive_attributes):
        err_msg_1 = "quasi identifiers are not in columns of the dataframe"
        err_msg_2 = "Sensitive Attributes are not in columns of the dataframe"
        err_msg_3 = "Sensitive Attributes cannot be a subset of quasi identifiers"
        err_msg_7 = "Categorical index list cannot be more than quasi identifiers"
        err_msg_4 = "Invalid Catergorical index"
        err_msg_5 = "Duplicate Value"
        err_msg_6 = "Expect argument as a list"
        columns = dataframe.columns

        if(type(quasi_identifiers) != list):
            raise AnonymizeError(message = err_msg_6+" in "+ 'quasi identifiers')
        elif(type(sensitive_attributes) != list):
            raise AnonymizeError(message = err_msg_6+" in "+ 'sensitive attributes')
        elif(len(set(quasi_identifiers)) != len(quasi_identifiers)):
            raise AnonymizeError(message = err_msg_5+" in "+'quasi identifiers')
        elif(len(set(sensitive_attributes)) != len(sensitive_attributes)):
            raise AnonymizeError(message = err_msg_5+" in "+ 'sensitive attributes')      
        elif not(set(quasi_identifiers).issubset(set(columns))):
            raise AnonymizeError(message = err_msg_1)
        elif not(set(sensitive_attributes).issubset(set(columns))):
            raise AnonymizeError(message = err_msg_2)
        elif(len(set(sensitive_attributes).intersection(set(quasi_identifiers))) > 0):
            raise AnonymizeError(message = err_msg_3)
        elif(len(sensitive_attributes) == 0):
            raise AnonymizeError(message = "Sensitive Attributes cannot be empty")
        return True

def validator(dataframe,QI_,SA_,CAT_INDEXES_,nan_replacement_int,nan_replacement_str):
    global QI_LEN,QI,_df,SA,IS_CAT,QI_RANGE_VAL,QI_RANGE_VAL,CAT_UNIQUE,NUM_COL,CAT_COL,_DEBUG,RANGE_FIX,CAT_INDEXES,NUM_COL_RANGE,CAT_COL_RANGE
    err_msg_1 = "quasi identifiers are not in columns of the dataframe"
    err_msg_2 = "Sensitive Attributes are not in columns of the dataframe"
    err_msg_3 = "Sensitive Attributes cannot be a subset of quasi identifiers"
    err_msg_7 = "Categorical index list cannot be more than quasi identifiers"
    err_msg_4 = "Invalid Catergorical index"
    err_msg_5 = "Duplicate Value"
    err_msg_6 = "Expect argument as a list"
    columns = dataframe.columns
    if(type(QI_) != list):
        raise AnonymizeError(message = err_msg_6+" in "+ 'quasi identifiers')
    elif(type(SA_) != list):
        raise AnonymizeError(message = err_msg_6+" in "+ 'sensitive attributes')
    elif(type(CAT_INDEXES_) != list):
        raise AnonymizeError(message = err_msg_6+" in "+ 'catergorical index')
    elif(len(set(QI_)) != len(QI_)):
        raise AnonymizeError(message = err_msg_5+" in "+'quasi identifiers')
    elif(len(set(SA_)) != len(SA_)):
        raise AnonymizeError(message = err_msg_5+" in "+ 'sensitive attributes')
      
    elif not(set(QI_).issubset(set(columns))):
        raise AnonymizeError(message = err_msg_1)
    elif not(set(SA_).issubset(set(columns))):
        raise AnonymizeError(message = err_msg_2)
    elif(len(set(SA_).intersection(set(QI_))) > 0):
        raise AnonymizeError(message = err_msg_3)
    else:
        try:
            CAT_INDEXES_ = list(map(int,CAT_INDEXES_))
        except:
            raise AnonymizeError(message = err_msg_4+"\n Index should be a integer")
        CAT_INDEXES_sorted = sorted(CAT_INDEXES_)
        if(len(CAT_INDEXES_) != 0):
            if not(0 <= CAT_INDEXES_sorted[-1] < len(QI_)):
              raise AnonymizeError(message = err_msg_4+"\ncatergorical index should start with zero and in between 0 and number of quasi identifiers")
            if not(0 <= CAT_INDEXES_sorted[0] < len(QI_)):
              raise AnonymizeError(message = err_msg_4+"\ncatergorical index should start with zero and in between 0 and number of quasi identifiers")
        if(len(CAT_INDEXES_) > len(QI_)):
            raise AnonymizeError(message = err_msg_7)
        elif(len(set(CAT_INDEXES_)) != len(CAT_INDEXES_)):
            raise AnonymizeError(message = err_msg_5 +" in "+'Categorical Index')
        IS_CAT = [False]*len(QI_)
        try:
            if(len(CAT_INDEXES_) != 0):
                for index in CAT_INDEXES_:
                    IS_CAT[index] = True
        except:
            raise AnonymizeError(message = "Invalid index for categorical indexes")
        NUM_COL_RANGE=[]
        CAT_COL_RANGE=[]
        QI = QI_
        SA = SA_
        QI_LEN,QI,_df,SA,IS_CAT,QI_RANGE_VAL,QI_RANGE_VAL,CAT_UNIQUE,NUM_COL,CAT_COL,_DEBUG,RANGE_FIX,CAT_INDEXES = marking_globals(dataframe)
        _df = df_validator(_df,nan_replacement_int,nan_replacement_str)
        for i in range(QI_LEN):
            if(IS_CAT[i] is False):
                diff = _df[QI[i]].max() - _df[QI[i]].min()
                NUM_COL_RANGE.append(diff)
            else:
                CAT_COL_RANGE.append(len(_df[QI[i]].unique()))
        return QI_LEN,QI,_df,SA,IS_CAT,QI_RANGE_VAL,QI_RANGE_VAL,CAT_UNIQUE,NUM_COL,CAT_COL,_DEBUG,RANGE_FIX,CAT_INDEXES,NUM_COL_RANGE,CAT_COL_RANGE

def df_validator(df,nan_replacement_int=0,nan_replacement_str=''):
    df = df.dropna(how='all')
    df[CAT_COL] = df[CAT_COL].applymap(str).fillna(nan_replacement_str)
    df[CAT_COL] = df[CAT_COL].applymap(lambda x: x.replace(",","/"))
    df[NUM_COL] = df[NUM_COL].fillna(nan_replacement_int).applymap(lambda x: numerical_validator(x,nan_replacement_int))
    return df


def numerical_validator(value,nan_replace_int):
    try:
        if(type(value) == int or type(value) == float):
            return value
        elif('-' in str(value) or '- ' in str(value) or ' -' in str(value)):
            elements = value.split('-')
            return (int(elements[0].strip()) + int(elements[1].strip()))//2
        else:
            return int(value)
    except ValueError:
        return nan_replace_int
    except Exception as e:
        if(_DEBUG):
            print(e)
        return nan_replace_int

def marking_globals(df):
    QI_LEN = len(QI)
    CAT_INDEXES,CAT_COL,NUM_COL = [],[],[]
    drop_col = []
    CAT_UNIQUE = []
    _DEBUG = False
    QI_RANGE_VAL = []
    RANGE_FIX = 1
    if(_DEBUG):
      print("Starting initializing globals")
    for column in df.columns:
        if not((column in QI) | (column in SA)):
            drop_col.append(column)
    if(_DEBUG):
      print("After initializing drop column")
    df = df.drop(drop_col, axis=1)
    # df = df[QI+SA]
    _df = df.loc[:,QI+SA]
    sensitive_input = df.loc[:,SA]
    if(_DEBUG):
      print("Before marking QI_RANGE_VAL")
    try:
      for i in range(QI_LEN):
          if(IS_CAT[i] is False):
              diff = df[QI[i]].max() - df[QI[i]].min()
              QI_RANGE_VAL.append(diff)
          else:
            unique_count = len(df[QI[i]].unique())
            CAT_UNIQUE.append(unique_count)
            QI_RANGE_VAL.append(unique_count)
    except:
      raise(AnonymizeError(message = "Invalid categorical index. Check whether catergorical indexes are correct"))
    if(_DEBUG):
      print("Before marking NUM_COL & CAT_COL & CAT_INDEXES")
    for i,ele in enumerate(IS_CAT):
        if(ele):
          CAT_COL.append(QI[i])
          CAT_INDEXES.append(i)
        else:
          NUM_COL.append(QI[i])
    if(_DEBUG):
        print("Finished initializing Globals")
    return QI_LEN,QI,_df,SA,IS_CAT,QI_RANGE_VAL,QI_RANGE_VAL,CAT_UNIQUE,NUM_COL,CAT_COL,_DEBUG,RANGE_FIX,CAT_INDEXES


class AnonymizeError(Exception):
    def __init__(self, message="Invalid Input"):
        self.message = message
        super().__init__(self.message)





# from pyspark.sql.functions import PandasUDFType, lit, pandas_udf
from .mondrian_utils.utility import *


def k_anonymizer(df, k, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    partitions = partition_dataset(
        df, k, None, None,  categorical, feature_columns, sensitive_column, full_spans)

    return anonymizer(df, partitions, feature_columns, sensitive_column, categorical)


def l_diversity_anonymizer(df, k, l, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    partitions = partition_dataset(
        df, k, l, None,  categorical, feature_columns, sensitive_column, full_spans)

    return anonymizer(df, partitions, feature_columns, sensitive_column, categorical)


def t_closeness_anonymizer(df, k, t, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    partitions = partition_dataset(
        df, k, None, t,  categorical, feature_columns, sensitive_column, full_spans)

    return anonymizer(df, partitions, feature_columns, sensitive_column, categorical)


def k_anonymizer_w_user(df, k, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    partitions = partition_dataset(
        df, k, None, None,  categorical, feature_columns, sensitive_column, full_spans)

    return anonymize_w_user(
        df, partitions, feature_columns, sensitive_column, categorical)


def l_diversity_anonymizer_w_user(df, k, l, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    partitions = partition_dataset(
        df, k, l, None,  categorical, feature_columns, sensitive_column, full_spans)

    return anonymize_w_user(
        df, partitions, feature_columns, sensitive_column, categorical)


def t_closeness_anonymizer_w_user(df, k, t, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    partitions = partition_dataset(
        df, k, None, t,  categorical, feature_columns, sensitive_column, full_spans)

    return anonymize_w_user(
        df, partitions, feature_columns, sensitive_column, categorical)


class Preserver:

    @staticmethod
    def k_anonymize(df, k, feature_columns, sensitive_column, categorical, schema):

        # def anonymize(pdf):
        a_df = k_anonymizer(df, k, feature_columns,
                            sensitive_column, categorical)
        return a_df

    @staticmethod
    def k_anonymize_w_user(df, k, feature_columns, sensitive_column, categorical, schema):
        a_df = k_anonymizer_w_user(df, k, feature_columns,
                                    sensitive_column, categorical)
        return a_df

    @staticmethod
    def l_diversity(df, k, l, feature_columns, sensitive_column, categorical, schema):
        a_df = l_diversity_anonymizer(df, k, l, feature_columns,
                                        sensitive_column, categorical)
        return a_df

    @staticmethod
    def l_diversity_w_user(df, k, l, feature_columns, sensitive_column, categorical, schema):
        a_df = l_diversity_anonymizer_w_user(df, k, l, feature_columns,
                                                sensitive_column, categorical)
        return a_df

    @staticmethod
    def t_closeness(df, k, t, feature_columns, sensitive_column, categorical, schema):
        a_df = t_closeness_anonymizer(df, k, t, feature_columns,
                                        sensitive_column, categorical)
        return a_df

    @staticmethod
    def t_closeness_w_user(df, k, t, feature_columns, sensitive_column, categorical, schema):
        a_df = t_closeness_anonymizer_w_user(df, k, t, feature_columns,
                                                sensitive_column, categorical)
        return a_df


    @staticmethod
    def anonymize_user(df, k, user, usercolumn_name, sensitive_column, categorical, schema, random=False):
        a_df = user_anonymizer(
            df, k, user, usercolumn_name, sensitive_column, categorical, random)
        return a_df



class Dataloss:

    @staticmethod
    def complete_data_loss(df,factor = 20):
        clusters = df.groupby('cluster_number')
        data_loss = clusters.apply(lambda cluster: cluster_data_loss(cluster)).sum(axis=1).sum()
        total_loss = factor*len(df)*len(NUM_COL) + 10*len(df)*len(CAT_COL)
        return data_loss/total_loss


def cluster_data_loss(cluster,factor=20):
    num_cols = cluster[NUM_COL]
    cat_cols = cluster[CAT_COL]
    num_loss = abs((num_cols.max()-num_cols.min())/NUM_COL_RANGE*factor).sum()
    cat_loss = cat_col_data_loss(cat_cols.loc[cat_cols.index[0]])*len(cluster)
    return cat_loss+num_loss

def cat_col_data_loss(column):
    temp = (column.str.split(',').str.len())/CAT_COL_RANGE*10
    return temp


import hashlib
import pandas as pd

# """Custom Error class"""


class AnonymizeError(Exception):
    def __init__(self, message):
        self.message = message


# """
# @PARAMS - get_spans()
# df - pandas dataframe
# partition - parition for whic to calculate the spans
# scale: if given, the spans of each column will be divided
#         by the scale for that column
# """


def get_spans(df, categorical, partition, scale=None):

    columns = list(df.columns)
    spans = {}
    for column in df.columns:
        if column in categorical:
            span = len(df[column][partition].unique())
        else:
            span = df[column][partition].max()-df[column][partition].min()
        if scale is not None:
            span = span/scale[column]
        spans[column] = span
    return spans


def get_full_span(df, categorical):
    for name in df.columns:
        if name not in categorical:
            df[name] = pd.to_numeric(df[name])

    return get_spans(df, categorical, df.index)


# """
# @PARAMS - split()
# df - pandas dataframe
# partition - parition for whic to calculate the spans
# column: column to split
# """


def split(df, categorical, partition, column):
    dfp = df[column][partition]
    if column in categorical:
        values = dfp.unique()
        lv = set(values[:len(values)//2])
        rv = set(values[len(values)//2:])
        return dfp.index[dfp.isin(lv)], dfp.index[dfp.isin(rv)]
    else:
        median = dfp.median()
        dfl = dfp.index[dfp < median]
        dfr = dfp.index[dfp >= median]
        return (dfl, dfr)


def is_k_anonymous(partition, k):

    if len(partition) < k:
        return False
    return True


def l_diversity(df, partition, column):
    return len(df[column][partition].unique())


def is_l_diverse(df, partition, sensitive_column, l):
    return l_diversity(df, partition, sensitive_column) >= l


# """
# @PARAMS - t_closeness()
# global_freqs: The global frequencies of the sensitive attribute values

# """


def t_closeness(df, partition, column, global_freqs):
    total_count = float(len(partition))
    d_max = None
    group_counts = df.loc[partition].groupby(column)[column].agg('count')
    for value, count in group_counts.to_dict().items():
        p = count/total_count
        d = abs(p-global_freqs[value])
        if d_max is None or d > d_max:
            d_max = d
    return d_max


# """
# @PARAMS - is_t_close()
# global_freqs: The global frequencies of the sensitive attribute values
# p: The maximum aloowed distance
# """


def is_t_close(df, partition, categorical, sensitive_column, global_freqs, p):

    if not sensitive_column in categorical:
        raise ValueError("T closeness is only for categorical values")
    result = t_closeness(df, partition, sensitive_column, global_freqs) <= p
    if(result):
        return result
    else:
        print("No T closseness")


def get_global_freq(df, sensitive_column):
    global_freqs = {}
    total_count = float(len(df))
    group_counts = df.groupby(sensitive_column)[sensitive_column].agg('count')

    for value, count in group_counts.to_dict().items():
        p = count/total_count
        global_freqs[value] = p
    return global_freqs


# @PARAMS - partition_dataset()
# df - pandas dataframe
# feature_column - list of column names along which to partitions the dataset
# scale - column spans


def partition_dataset(df, k, l, t, categorical, feature_columns, sensitive_column, scale):

    finished_partitions = []
    global_freqs = {}
    if t is not None:
        global_freqs = get_global_freq(df, sensitive_column)

    partitions = [df.index]
    while partitions:
        partition = partitions.pop(0)
        spans = get_spans(df[feature_columns],
                          categorical, partition, scale)
        for column, span in sorted(spans.items(), key=lambda x: -x[1]):
            lp, rp = split(df, categorical, partition, column)
            if l is not None:
                if not is_k_anonymous(lp, k) or not is_k_anonymous(rp, k) or not is_l_diverse(df, lp, sensitive_column, l) or not is_l_diverse(df, rp, sensitive_column, l):
                    continue
            if l is None:
                if t is None:
                    if not is_k_anonymous(lp, k) or not is_k_anonymous(rp, k):
                        continue
                if t is not None:
                    if not is_k_anonymous(lp, k) or not is_k_anonymous(rp, k) or not is_t_close(df, lp, categorical, sensitive_column, global_freqs, t) or not is_t_close(df, rp, categorical, sensitive_column, global_freqs, t):
                        continue
            partitions.extend((lp, rp))
            break
        else:
            finished_partitions.append(partition)
    return finished_partitions


def agg_categorical_column(series):
    return ','.join(sorted(set(series)))


def agg_numerical_column(series):
    minimum = series.min()
    maximum = series.max()
    if(maximum == minimum):
        string = str(maximum)
    else:
        string = ''
        maxm = str(maximum)
        minm = str(minimum)

        if(len(minm) == 1):
            if(minimum >= 5):
                string = '5-'
            else:
                string = '0-'
        else:
            if (minm[-1]=='0'):
                string = minm +"-"
            else:
                min_start = minm[:-1]
                if(minimum >= int(min_start+'5')):
                    string = min_start+'5-'
                else:
                    string = min_start+'0-'

        if(len(maxm) == 1):
            if(maximum >= 5):
                string += "10"
            else:
                string += '5'
        else:
            if(maxm[-1]=='0'):
                string += maxm
            else:
                max_start = maxm[:-1]
                if(maximum > int(max_start+'5')):
                    string += str(int(max_start+'0') + 10)
                else:
                    string += max_start+'5'

    return string

def anonymizer(df, partitions, feature_columns, sensitive_column, categorical, max_partitions=None):
    aggregations = {}

    for column in feature_columns:
        if column in categorical:
            aggregations[column] = agg_categorical_column
        else:
            aggregations[column] = agg_numerical_column
    rows = []

    for i, partition in enumerate(partitions):
        if i % 100 == 1:
            print("Finished processing {} partitions.".format(i))
        if max_partitions is not None and i > max_partitions:
            break

        sortby = ['_common88column_'] + list(categorical)
        grouped_columns = df.loc[partition].assign(_common88column_=1).sort_values(
            by=sortby,ascending=False).groupby(
                '_common88column_', sort=False).agg(aggregations, squeeze=False)
        sensitive_counts = df.loc[partition].groupby(
            sensitive_column).agg({sensitive_column: 'count'})
        values = grouped_columns.iloc[0].to_dict()
        for sensitive_value, count in sensitive_counts[sensitive_column].items():
            if count == 0:
                continue
            values.update({
                sensitive_column: sensitive_value,
                'count': count,

            })
            rows.append(values.copy())
    dfn = pd.DataFrame(rows)
    pdfn = dfn.sort_values(feature_columns+[sensitive_column])
    return pdfn


# """ --------------------------------------------------------------------------
    # Single User Anonymize
# """ --------------------------------------------------------------------------


def getIntersection(df, udf, user, threshold, columns, usercolumn_name):
    i = 0
    intersect_df = pd.DataFrame()
    for column in columns:
        i += 1
        if (i > threshold):
            break
        val = udf[column].value_counts().idxmax()
        tempdf = df.loc[(df[column] == val) & (df[usercolumn_name] != user)]
        if (intersect_df.empty):
            intersect_df = tempdf
        if(not tempdf.empty):
            intersect_df = tempdf.reset_index().merge(
                intersect_df, how='inner').set_index('index')

    return intersect_df


def commonDF(df, udf, user, requiredRows, columns, usercolumn_name, random):
    length = len(columns)
    for i in range(length):
        intersect_df = getIntersection(
            df, udf, user, length-i, columns, usercolumn_name)
        if(intersect_df.shape[0] >= requiredRows):
            break
        else:
            continue

    revcolumn = columns[::-1]

    for i in range(length):
        intersect_df = getIntersection(
            df, udf, user, length-i, revcolumn, usercolumn_name)
        if(intersect_df.shape[0] >= requiredRows):
            break
        else:
            continue

    if (intersect_df.shape[0] < requiredRows):
        for column in columns:
            intersect_df = getIntersection(
                df, udf, user, 1, [column], usercolumn_name)
            if(intersect_df.shape[0] >= requiredRows):
                break
    if ((intersect_df.shape[0] < requiredRows) & random):
        try:
            intersect_df = df.sample(requiredRows)
        except ValueError:
            raise(AnonymizeError("Data frame is not enough for anonymization"))
    return intersect_df


def anonymizeGivenUser(df, udf, user, usercolumn_name, columns, categorical):
    indexes = list(udf.index)
    for column in columns:
        if column not in categorical:
            udf[column] = pd.to_numeric(udf[column])
            df[column] = pd.to_numeric(df[column])
        valueList = udf[column].unique()
        if column in categorical:
            string = ','.join(valueList)
            df[column] = df[column].astype(str)
            df.loc[indexes, column] = string
        if column not in categorical:

            minimum = min(valueList)
            maximum = max(valueList)
            if(maximum == minimum):
                string = str(maximum)
            else:
                string = ''
                maxm = str(maximum)
                minm = str(minimum)

                if(len(minm) == 1):
                    min_start = minm[-1]
                    if(minimum >= 5):
                        string = '5-'
                    else:
                        string = '0-'
                else:
                    min_start = minm[-2]
                    if(minimum >= int(min_start+'5')):
                        string = min_start+'5-'
                    else:
                        string = min_start+'0-'

                if(len(maxm) == 1):
                    max_start = maxm[-1]
                    if(maximum >= 5):
                        string += "10"
                    else:
                        string += '5'
                else:
                    max_start = maxm[-2]
                    if(maximum >= int(max_start+'5')):
                        string += str(int(max_start+'0') + 10)
                    else:
                        string += max_start+'5'

                        min_start = minm[-2]
                        max_start = maxm[-2]

            df[column] = df[column].astype(str)
            df.loc[indexes, column] = string


def user_anonymizer(df, k, user, usercolumn_name, sensitive_column, categorical, random=False):

    if ((sensitive_column not in df.columns) or (usercolumn_name not in df.columns)):
        raise AnonymizeError("No Such Sensitive Column")

    df[usercolumn_name] = df[usercolumn_name].astype(str)

    userdf = df.loc[df[usercolumn_name] == str(user)]
    user = str(user)
    if(userdf.empty):
        raise AnonymizeError("No user found.")

    rowcount = userdf.shape[0]
    columns = userdf.columns.drop([usercolumn_name, sensitive_column])

    if (rowcount >= k):
        requiredRows = 1
    else:
        requiredRows = k - rowcount
    intersect_df = commonDF(
        df, userdf, user, requiredRows, columns, usercolumn_name, random)

    if((not intersect_df.empty) & (intersect_df.shape[0] >= requiredRows)):
        finaldf = pd.concat([userdf, intersect_df])
        anonymizeGivenUser(df, finaldf, user,
                           usercolumn_name, columns, categorical)
    else:
        raise(AnonymizeError("Can't K Anonymize the user for given K value"))
    return df


# """ --------------------------------------------------------------------------
    # Anonymize with all rows
# """ --------------------------------------------------------------------------
def agg_columns(df, partdf, indexes, columns, categorical):

    for column in columns:

        if column not in categorical:
            partdf[column] = pd.to_numeric(partdf[column])
        valueList = partdf[column].unique()

        if column in categorical:
            string = ','.join(valueList)
            df[column] = df[column].astype(str)
            df.loc[indexes, column] = string

        if column not in categorical:
            minimum = min(valueList)
            maximum = max(valueList)
            if(maximum == minimum):
                string = str(maximum)
            else:
                string = ''
                maxm = str(maximum)
                minm = str(minimum)
                if(len(minm) == 1):
                    if(minimum >= 5):
                        string = '5-'
                    else:
                        string = '0-'
                else:
                    if (minm[-1]=='0'):
                        string = minm +"-"
                    else:
                        min_start = minm[:-1]
                        if(minimum >= int(min_start+'5')):
                            string = min_start+'5-'
                        else:
                            string = min_start+'0-'

                if(len(maxm) == 1):
                    if(maximum >= 5):
                        string += "10"
                    else:
                        string += '5'
                else:
                    if(maxm[-1]=='0'):
                        string += maxm
                    else:
                        max_start = maxm[:-1]
                        if(maximum > int(max_start+'5')):
                            string += str(int(max_start+'0') + 10)
                        else:
                            string += max_start+'5'

            df[column] = df[column].astype(str)
            df.loc[indexes, column] = string


def anonymize_w_user(df, partitions, feature_columns, sensitive_column, categorical):

    if sensitive_column not in df.columns:
        raise AnonymizeError("No Such Sensitive Column")

    for fcolumn in feature_columns:
        if fcolumn not in df.columns:
            raise AnonymizeError("No Such Feature Column :"+fcolumn)

    full_spans = get_full_span(df, categorical)
    aggregations = {}
    df_copy = df.copy()
    for column in feature_columns:
        if column in categorical:
            aggregations[column] = agg_categorical_column
        else:
            aggregations[column] = agg_numerical_column

    for i, partition in enumerate(partitions):
        if i % 100 == 1:
            print("Finished processing {} partitions.".format(i))

        partdf = df.loc[partition]
        agg_columns(df, partdf, partition, feature_columns, categorical)

    df = df.sort_values(feature_columns+[sensitive_column])
    return df



import numpy as np
from .. import gv

class Calculator:

    @staticmethod
    def cal_num_col_dist(record,cluster_record,range_value,factor):
        '''
        This method is used to find the numerical distance between records and cluster genaralization values
        record :-  The record need to find the distance with cluster  type :- Pandas Series
        cluster_record :- Numerical generalization values of the clusters   type :- Pandas DataFrame
        range_value :- Clusters range values   type :- Pandas DataFrame
        factor :- factor need to multiply. Recommended 20. type:- Integer
        '''
        range_value = range_value.where(range_value != 0, np.nan)
        return abs((cluster_record - record)/(range_value*factor)).fillna(0)

    @staticmethod
    def cal_cat_col_dist1(row,categorical_col):
        '''
        This method is used to find the categorical distance between records and cluster genaralization values
        row :-  The record need to find the distance with cluster  type :- Pandas Series
        categorical_col :- Catergorical generalization values of the clusters   type :- Pandas DataFrame
        '''
        return (len(gv.GV['CAT_COL'])-np.where(
            row[gv.GV['CAT_COL']]==categorical_col,True,np.where(
                categorical_col == "*****",True,False)).sum(axis=1))*10
    @staticmethod
    def cal_cat_col_dist2(row,categorical_col):
        '''
        This method is used to find the categorical distance between records and cluster genaralization values
        row :-  The record need to find the distance with cluster  type :- Pandas Series
        categorical_col :- Catergorical generalization values of the clusters   type :- Pandas DataFrame
        '''
        return (len(gv.GV['CAT_COL'])-np.where(
            row[gv.GV['CAT_COL']]==categorical_col,1,np.where(
                categorical_col == "*****",0.5,0)).sum(axis=1))*10

    @staticmethod
    def cal_cat_col_dist3(row,categorical_col):
        '''
        This method is used to find the categorical distance between records and cluster genaralization values
        row :-  The record need to find the distance with cluster  type :- Pandas Series
        categorical_col :- Catergorical generalization values of the clusters   type :- Pandas DataFrame
        '''
        return (len(gv.GV['CAT_COL'])-np.where(
            row[gv.GV['CAT_COL']]==categorical_col,1,0).sum(axis=1))*10

    @staticmethod
    def numerical_dataloss(cluster,cluster_list,range_value,factor=20):
        '''
        This function used to find the numerical dataloss in a situation of two clusters join
        cluster :- Numerical values in the cluster type:-Pandas DataFrameGroupBy
        cluster_list :- Numerical values in the groups of clusters  type:-Pandas DataFrameGroupBy
        range_value :- Ranges for numerical values type :- Pandas DataFrame
        factor :- factor need to multiply. Recommended 20. type:- Integer
        '''
        return np.sum(((abs(cluster_list.min() - cluster.min()))+abs(cluster_list.max() - cluster.max()))/range_value*factor,axis=1)

    @staticmethod
    def categorical_dataloss(cluster,cluster_list,factor=20):
        '''
        This method is used to compare unique catergorical values in two clusters.
        cluster :- Categorical values in the cluster. type :- numpy array
        cluster_list :- Catergorical values in cluster values. type :- Pandas numpy array
        factor :- Factor to multiply the dataloss. Recomonded 20. Type Integer
        '''
        cluster = cluster.reshape((1,cluster.shape[0]))
        cluster_list = cluster_list.reshape((1,cluster_list.shape[0]))
        intersection_list = np.intersect1d(cluster[0],cluster_list[0])
        return abs(cluster[0].size+cluster_list[0].size-2*intersection_list.size)


# from pyspark.sql.functions import udf  # type: ignore
# from pyspark.sql.types import DoubleType, StringType  # type: ignore
from diffprivlib.mechanisms import LaplaceTruncated, Binary  # type: ignore
from tabulate import tabulate  # type: ignore
from pandas import DataFrame
# Following imports provide type checking functionality to the library
# from pyspark.sql.dataframe import DataFrame as SparkDataFrame  # type: ignore
from numbers import Real  # type: ignore
from typing import List, Dict, Union, Optional  # type: ignore
import sys  # type: ignore

if sys.version_info >= (3, 8):
    from typing import TypedDict  # type: ignore
else:
    from typing_extensions import TypedDict  # type: ignore


class Column(TypedDict, total=False):
    column_name: str
    category: str
    epsilon: Union[int, float]
    delta: Union[int, float]
    sensitivity: Union[int, float, None]
    lower_bound: Union[int, float]
    upper_bound: Union[int, float]
    round: int
    label1: Optional[str]
    label2: Optional[str]


class DPLib:
    r""" Create a differentially private Spark DataFrame from an existing Spark DataFrame

    This class makes use of the 'diffprivlib' library by IBM to create differentially private Spark DataFrame.
    It can handle columns with two major data types: 'numeric' and 'boolean'
        'numeric': column should only have numbers or 'null' value
                    utilizes LaplaceTruncated mechanism from diffprivlib.mechanisms

        'boolean': each row of column should have one of two boolean values set by user
                    utilizes Binary mechanism from diffprivlib.mechanisms

    Attributes:
        sdf: A Spark DataFrame object. Methods: `__check_labels`, `set_column` and `execute`
            will raise ValueError if sdf is None.

    Public methods:
        set_global_epsilon_delta:
            Assigns common epsilon and delta to be used by all columns if they lack
            column specific values for epsilon and delta

        set_global_sensitivity:
            Assigns common sensitivity to be used by all columns if they lack
            column specific value for sensitivity

        set_sdf: Assigns DataFrame object to the class

        set_column: Adds column dictionary to `__columns`

        execute: changes existing sdf to be differentially private. This change is not reversible.

    Examples:
        Check out `differential_preserver_demo.ipynb` Jupyter notebook

    """

    def __init__(self,
                 global_epsilon: Union[int, float, None] = None,
                 global_delta: Union[int, float] = 0.0,
                 sdf: Optional[DataFrame] = None) -> None:
        r""" Inits DPLib

        Args:
            global_epsilon: Common epsilon value to be used by all columns as a fail-safe
            global_delta: Common delta value to be used by all columns as a fail-safe. Defaults to 0.0
            sdf: Spark DataFrame to be converted. Can change to a different DataFrame with set_df()
        """

        self.sdf: Optional[DataFrame] = None
        self.set_sdf(sdf)

        self.__columns: Dict[str, Column] = {}

        self.__epsilon: Optional[float] = None
        self.__delta: Optional[float] = None
        self.__sensitivity: Union[int, float, None] = None

        if global_epsilon is not None and self.__check_epsilon_delta(global_epsilon, global_delta):
            self.__epsilon = float(global_epsilon)
            self.__delta = float(global_delta)

    @staticmethod
    def __check_epsilon_delta(epsilon: Union[int, float], delta: Union[int, float]) -> bool:
        r""" checks whether epsilon and delta meet required conditions for method `execute`

        Always called disregarding category. Both 'LaplaceTruncated' and 'Binary' mechanism
        require epsilon and delta.

        Args:
            epsilon: epsilon value to be used by method `execute`
            delta: delta value to be used by method `execute`

        Returns: True if parameters satisfy the conditions

        Raises: TypeError, ValueError if parameters do not obey the rules

        """
        if not isinstance(epsilon, Real) or not isinstance(delta, Real):
            raise TypeError("Epsilon and delta must be numeric")

        if epsilon < 0:
            raise ValueError("Epsilon must be non-negative")

        if not 0 <= float(delta) <= 1:
            raise ValueError("Delta must be in range [0, 1]")

        if epsilon == 0 and delta == 0:
            raise ValueError("Epsilon and Delta cannot both be zero")

        return True

    @staticmethod
    def __check_sensitivity(sensitivity: Union[int, float, None]) -> bool:
        r""" checks whether sensitivity meets required conditions for method `execute`

        Called only when category = 'numeric'

        Args:
            sensitivity: sensitivity value to be used by method `execute`.

        Returns: True if parameter satisfies the conditions

        Raises: TypeError, ValueError if parameter does not obey the rules

        """
        if not isinstance(sensitivity, Real):
            raise TypeError("Sensitivity must be numeric")

        if sensitivity <= 0:
            raise ValueError("Sensitivity must be strictly positive")

        return True

    @staticmethod
    def __check_labels(sdf: DataFrame, column_name: str, label1: Optional[str], label2: Optional[str]) -> bool:
        r""" checks whether labels meet required conditions for method `execute`.

        Called only when category = 'boolean'

        Args:
            sdf: Spark DataFrame object
            column_name: specific column which is to be executed using 'Binary' mechanism
            label1: label to be used by 'Binary' mechanism
            label2: label to be used by 'Binary' mechanism

        Returns: True if parameters satisfy the conditions

        Raises: TypeError, ValueError if parameters do not obey the rules

        """
        if not isinstance(label1, str) or not isinstance(label2, str):
            raise TypeError("Labels must be strings.")

        if len(label1) == 0 or len(label2) == 0:
            raise ValueError("Labels must be non-empty strings")

        if label1 == label2:
            raise ValueError("Labels must not match")

        # finds unique values in a column
        labels: List[str] = [row[column_name] for row in sdf.select(column_name).distinct().collect()]

        if len(labels) != 2 or label1 not in labels or label2 not in labels:
            # checks whether all the rows of column have either label1 or label2
            raise ValueError("Labels in column `%s` does not match with labels entered" % column_name)

        return True

    @staticmethod
    def __check_bounds(lower: Union[int, float], upper: Union[int, float]) -> bool:
        r""" checks whether lower and upper bounds for a specific column bound to their conditions

        Called only when category = 'numeric'

        Args:
            lower: lower bound of a column
            upper: upper bound of a column

        Returns: True if parameters satisfy the conditions

        Raises: TypeError, ValueError if parameters do not obey the rules

        """

        if not isinstance(lower, Real) or not isinstance(upper, Real):
            raise TypeError("Bounds must be numeric")

        if lower > upper:
            raise ValueError("Lower bound must not be greater than upper bound")

        return True

    @staticmethod
    def __check_sdf(sdf: DataFrame) -> bool:
        r"""

        Args:
            sdf: Spark DataFrame

        Returns: True if sdf satisfies the condition

        Raises: TypeError if sdf does not obey the rules

        """

        if sdf is not None and not isinstance(sdf, DataFrame):
            raise TypeError('Given sdf is not a Spark DataFrame')

        return True

    def set_global_epsilon_delta(self, epsilon: Union[int, float], delta: Union[int, float] = 0.0) -> None:
        r"""  set global epsilon and delta if they satisfy method `__check_epsilon_delta`

        Args:
            epsilon: epsilon value to be used by method `execute`
            delta: delta value to be used by method `execute`

        Raises:
            The method itself will not raise any exceptions.
            However inner method `__check_epsilon_delta` may raise exceptions
        """

        if self.__check_epsilon_delta(epsilon, delta):
            self.__epsilon = float(epsilon)
            self.__delta = float(delta)

    def set_global_sensitivity(self, sensitivity: Union[int, float]) -> None:
        r"""  set global sensitivity if it satisfies method `__check_sensitivity`

        Args:
            sensitivity: sensitivity value to be used by method `execute`.

        Raises:
            The method itself will not raise any exceptions.
            However inner method `__check_sensitivity` may raise exceptions
        """

        if self.__check_sensitivity(sensitivity):
            self.__sensitivity = float(sensitivity)

    def set_sdf(self, sdf: DataFrame) -> None:
        r""" sets Spark DataFrame object to class.

        Args:
            sdf: A Spark DataFrame object

        """

        if self.__check_sdf(sdf):
            self.sdf = sdf

    def set_column(self, column_name: str, category: str,
                   epsilon: Union[int, float, None] = None, delta: Union[int, float, None] = None,
                   sensitivity: Union[int, float, None] = None,
                   lower_bound: Union[int, float, None] = None, upper_bound: Union[int, float, None] = None,
                   round: Optional[int] = None,
                   label1: Optional[str] = None, label2: Optional[str] = None) -> None:

        r""" adds a column with custom parameters to the __columns dictionary.

        A column may have specific details. Hence this method allows user to set them individually for a column.
        However in case any one parameter is missing, appropriate value will be copied from available global
        parameters and passed instead.

        Parameters: column_name, category are compulsory. However exceptions may arise when certain optional
        parameters such as lower_bound, upper_bound etc. are not set.

        Args:
            ----------------------- common arguments -----------------------------------------------
            column_name: Name of column
            category: Category the column belongs to: ['numeric', 'boolean']
            epsilon: Epsilon value to be used by method `execute`
            delta: Delta value to be used by method `execute`

            ----------------------- arguments specific to category = 'numeric' ----------------------
            sensitivity: Sensitivity value to be used by method `execute`.
            lower_bound: Lower bound of a column.
            upper_bound: Upper bound of a column.
            round: Rounding factor. Values can be rounded off after applying a certain mechanism.

            ----------------------- arguments specific to category = 'boolean' ----------------------
            label1: label to be used by 'Binary' mechanism.
            label2: label to be used by 'Binary' mechanism.

        Raises: TypeError, ValueError if parameters have not been set correctly.
                Inner methods may raise Exception.

        """

        if self.sdf is None:
            raise ValueError("Add an eligible Spark DataFrame before adding columns")
        self.__check_sdf(self.sdf)

        if column_name not in self.sdf.columns:
            raise ValueError("Cannot find column in given DataFrame")

        if category not in ['numeric', 'boolean']:
            raise ValueError("category must be either `numeric` or `boolean`")

        column: Column = {'category': category}

        if epsilon is None:
            if self.__epsilon is not None:
                epsilon = self.__epsilon
            else:
                raise ValueError("Epsilon must be set")

        if delta is None:
            if self.__delta is not None:
                delta = self.__delta
            else:
                delta = 0.0

        if epsilon is not None and delta is not None and self.__check_epsilon_delta(epsilon, delta):
            column['epsilon'] = epsilon
            column['delta'] = delta

        if category == 'numeric':

            if sensitivity is None:
                sensitivity = self.__sensitivity
            if self.__check_sensitivity(sensitivity):
                column['sensitivity'] = sensitivity

            if lower_bound is None:
                lower_bound = float('-inf')
            if upper_bound is None:
                upper_bound = float('inf')
            if self.__check_bounds(lower_bound, upper_bound):
                column['lower_bound'] = lower_bound
                column['upper_bound'] = upper_bound

            if round is not None:
                if not isinstance(round, int) or round < 0:
                    raise TypeError("round must be positive integer")
                else:
                    column['round'] = round

        if category == 'boolean':
            if self.__check_labels(self.sdf, column_name, label1, label2):
                column['label1'] = label1
                column['label2'] = label2

        self.__columns[str(column_name)] = column

    def drop_column(self, *columns: str) -> None:
        """ drops a column added using `add_column()` method from `self.__columns`

        Args:
            *columns: one or more column names in string format separated by comma
                        if '*' is given, all columns will be dropped

        """

        if len(columns) == 1 and '*' in columns:
            self.__columns = {}
        else:
            for column in columns:
                if column in self.__columns:
                    del self.__columns[column]

    def get_config(self) -> None:

        print('Global parameters')
        print('-----------------\n')

        global_table: List[List[Union[float, str]]] = [
            ['Epsilon', self.__epsilon if self.__epsilon is not None else '--'],
            ['Delta', self.__delta if self.__delta is not None else '--'],
            ['Sensitivity', self.__sensitivity if self.__sensitivity is not None else '--']
        ]

        print(tabulate(tabular_data=global_table,
                       tablefmt='plain',
                       disable_numparse=True))

        print('\n')
        print('Column specific parameters')
        print('--------------------------\n')

        column_table: List[List] = []

        for column_name, details in self.__columns.items():
            row: List = [column_name,
                         details['category'],
                         details['epsilon'] if 'epsilon' in details else '--',
                         details['delta'] if 'delta' in details else '--',
                         details['sensitivity'] if 'sensitivity' in details else '--',
                         details['lower_bound'] if 'lower_bound' in details else '--',
                         details['upper_bound'] if 'upper_bound' in details else '--',
                         details['round'] if 'round' in details else '--',
                         details['label1'] if 'label1' in details else '--',
                         details['label2'] if 'label2' in details else '--'
                         ]
            column_table.append(row)

        if self.sdf is not None and isinstance(self.sdf, DataFrame):
            for column in self.sdf.columns:
                if column not in self.__columns:
                    row = [column]
                    row += '--' * 9
                    column_table.append(row)

        print(tabulate(tabular_data=column_table,
                       tablefmt='github',
                       disable_numparse=True,
                       headers=['Column name', 'Column category', 'Epsilon', 'Delta', 'Sensitivity', 'Lower bound',
                                'Upper bound', 'Round', 'Label 1', 'Label 2']
                       )
              )

    def execute(self):
        r"""

        Raises:
            ValueError: if self.__columns is empty
            inner methods may raise Exception, only if parameters have not been set correctly

        """

        laplace: LaplaceTruncated = LaplaceTruncated()
        binary: Binary = Binary()

        if self.__columns == {}:
            raise ValueError('No columns added for execution')

        for column_name, details in self.__columns.items():

            if details['category'] == 'numeric':

                # self.sdf = self.sdf.withColumn(colName=column_name,
                #                                col=self.sdf[column_name].cast(DoubleType()))
                self.sdf = self.sdf.assign({column_name : lambda x: 0})

                laplace.set_epsilon_delta(epsilon=details['epsilon'], delta=details['delta'])
                laplace.set_sensitivity(details['sensitivity'])
                laplace.set_bounds(lower=details['lower_bound'], upper=details['upper_bound'])

                if 'round' in details:

                    def round_randomise(cell):
                        return float(round(laplace.randomise(cell), details['round'])) if cell is not None else None


                    # self.sdf = self.sdf.withColumn(colName=column_name,
                    #                                col=round_randomise_udf(column_name))
                    self.sdf = self.sdf.assign({column_name: round_randomise})

                else:

                    def randomise(cell):
                        return float(laplace.randomise(cell)) if cell is not None else None

                    # randomise_udf = udf(f=randomise, returnType=DoubleType())

                    # self.sdf = self.sdf.withColumn(colName=column_name,
                    #                                col=randomise_udf(column_name))
                    self.sdf = self.sdf.assign({column_name: round_randomise})

            elif details['category'] == 'boolean':

                # self.sdf = self.sdf.withColumn(colName=column_name,
                #                                col=self.sdf[column_name].cast(StringType()))

                binary.set_epsilon_delta(epsilon=details['epsilon'], delta=details['delta'])
                binary.set_labels(value0=details['label1'], value1=details['label2'])

                def randomise(cell):
                    return binary.randomise(cell) if cell is not None else None

                # randomise_udf = udf(f=randomise, returnType=StringType())

                # self.sdf = self.sdf.withColumn(colName=column_name,
                #                                col=randomise_udf(column_name))
                self.sdf = self.sdf.assign({column_name: randomise})




from .. import gv
import numpy as np
class ClusterInit:

    @staticmethod
    def find_best_cluster_gens(n_clusters,dataframe):
        
        length = len(dataframe)
        probs = [dataframe[column].value_counts()/length  for column in gv.GV['QI']]
        dataframe['probability_sum'] = dataframe[gv.GV['QI']].apply(
          lambda row: np.sum([probs[i][row[column]] for i,column in enumerate(gv.GV['QI'])]),axis = 1)
        points = dataframe.sample(n = n_clusters , weights = 'probability_sum')
        dataframe = dataframe[gv.GV['QI']]
        return points[gv.GV['QI']]

    @staticmethod
    def select_centroids_using_weighted_column(n_clusters,dataframe):
        counts = [len(dataframe[column].unique())  for column in gv.GV['QI']]
        column_counts = sorted(counts)
        column = counts.index(column_counts[-1])
        try:
          return dataframe.sample(n = n_clusters , weights = gv.GV['QI'][column])
        except Exception:
          return ClusterInit.find_best_cluster_gens(n_clusters,dataframe)
        

    @staticmethod
    def random_sample_centroids(n_clusters,unique_rows):
        return unique_rows.sample(n=n_clusters)


import numpy as np
import sys
import logging

logger = logging.getLogger(__name__)

logger.level = logging.DEBUG
stream_handler = logging.StreamHandler(sys.stderr)
logger.addHandler(stream_handler)


class Kmodehelpers:

    @staticmethod
    def edit_cluster(group,distances):
        '''
        This method is used to edit the cluster number of a record.
        group :- Cluster group, type = Pandas DataFrameGroupBy
        distances :- Dataframe that include cluster mapping details
        '''
        c_no = group['cluster_number'].unique()[0]
        group['cluster_number'] = distances[c_no]
        return group

    @staticmethod
    def numeric_range(numeric_values):
        '''
        This method is used to generalize the data in a cluster.
        The given values in a cluster turn to the min-max range if they are not same.
        numeric_values :- Numeric values of the cluster. type = Pandas DataFrame
        '''
        if numeric_values.empty:
            return numeric_values

        numerical_min = numeric_values.min()
        numerical_max = numeric_values.max()
        return numerical_min.where(numerical_min.eq(numerical_max),numerical_min.apply(str) +"-"+numerical_max.apply(str))

    @staticmethod
    def catergorical_range(catergorical):
        '''
        This method is used to anonymize catergorical values in a cluster.
        Given categorical values turn to comma seperated generalization if they are not same.
        catergorical :- Catergorical values in a cluster. type = Pandas DataFrame
        '''
        catergorical = catergorical.apply(str)
        return ','.join(catergorical.unique())

    # def remove_far_clusters(self):
    #     '''
    #     This method is used to remove the far away clusters according to users preference
    #     Class attribute anonimize_ratio used to measure the distance
    #     '''
    #     temp = self.df['cluster_number'].value_counts()
    #     very_small_clusters = temp.loc[temp <= self.k*self.anonimize_ratio].index
    #     not_small_cluster = self.df.apply(lambda row: row['cluster_number'] not in very_small_clusters,axis=1)
    #     self.centroids = self.centroids.iloc[temp.loc[temp > self.k*self.anonimize_ratio].index]
    #     self.df = self.df.loc[not_small_cluster]
    #     if(_DEBUG):
    #         print(self.df)
    #         print(self.centroids)
    #     return self.df.sort_values(by=['cluster_number']),self.centroids


    # def adjust_kless_clusters(self,n = 2,method='dataloss'):
    #     '''
    #     This method is used to anonymize clusters that have less than k members.
    #     n :- Define the number of clusters that will join with a single less member cluster. Type :- Integer
    #     method :- If this is dataloss, Distance of two clusters is measured by the dataloss of joining, else Distance between cluster centroids. Type :- String
    #     '''
    #     self.mark_less_n_kcentroids()
    #     if(_DEBUG):
    #       print("marked_less clusters")
    #     self.df_copy = self.df.copy()
    #     self.df_copy['is_anonimized'] = 0
    #     if(method == 'dataloss'):
    #         centroid_distances = self.cluster_data_loss()
    #     else: 
    #         centroid_distances = self.less_centroids.apply(lambda row : self.get_distance_centers(row),axis=1)
    #     if(_DEBUG):
    #       print("Get cenr distance")
    #     n_close_centroids = np.argsort(centroid_distances,axis=1).iloc[:,:n]
    #     if(_DEBUG):
    #       print("After taking closing")
    #     results =  self.less_centroids.apply(lambda row: self.anonimize_k_less_clusters(row,n_close_centroids),axis=1)
    #     return results

    # def anonimize_k_less_clusters(self,centroid,n_close_centroids):
    #     '''
    #     This method used to anonymize a single less member cluster
    #     centroid:- The centroid of the cluster. Type :- Pandas Series
    #     n_close_centroids :- This have the details about distances between cluster.(Distances measure by dataloss or centroid distances). Type :- Pandas DataFrame        
    #     '''
    #     if(_DEBUG):
    #       print("anonimize_k_less_clusters")
    #     close_cetroids = n_close_centroids.loc[centroid.name]
    #     if(_DEBUG):
    #       print("take_n_close")
    #     centroid_related_rows = self.df.loc[self.df.apply(lambda x:x['cluster_number'] in close_cetroids.values,axis=1)].sort_values(by=['cluster_number'])
    #     if(_DEBUG):
    #       print("centroid_related_rows")
    #     related_indexes = self.df_copy.apply(lambda x: x['cluster_number'] == centroid.name,axis=1)
    #     if(_DEBUG):
    #       print("related_indexes")
    #     numerical_min = centroid_related_rows[NUM_COL].min()
    #     numerical_max = centroid_related_rows[NUM_COL].max()
    #     centroid_related_rows[NUM_COL] = np.where(numerical_min == numerical_max,numerical_min,numerical_min.apply(str) +"-"+numerical_max.apply(str))
    #     centroid_related_rows[CAT_COL] = np.where(centroid_related_rows[CAT_COL].nunique() == 1,centroid_related_rows[CAT_COL],centroid_related_rows[CAT_COL].apply(anonimize_catergorical_value_in_cluster))
    #     centroid_related_rows = centroid_related_rows.loc[related_indexes]
    #     if(_DEBUG):
    #       print("centroid_related_rows")
    #     self.df_copy.at[related_indexes,'is_anonimized'] = 1
    #     self.df_copy.at[related_indexes,QI] = centroid_related_rows[QI]
    #     if(_DEBUG):
    #       print("above to finish k less")
    #     return centroid_related_rows





import pandas as pd
import numpy as np
import time
import datetime
import random
from . import gv
from kmodes.kmodes import KModes
# from pyspark.sql.functions import PandasUDFType, lit, pandas_udf
from .clustering_utils.input_validate import InputValidator
# import clustering_utils.input_validate.InputValidator

from .clustering_utils.data_loss import Dataloss
from .clustering_utils.clustering import Clustering
from .clustering_utils.cluster_init import ClusterInit
from .clustering_utils.kmodes import Kmodehelpers
from .clustering_utils.distance_calculation import Calculator

class Kanonymizer(object):
    def __init__(self, df, QI_attr, Sensitive_attr, cat_indecies, verbose=1, max_iter=10, anonimize_ratio=1, max_cluster_distance=20):
        '''

        Attributes:-
          df :- Pandas Dataframe that will used to anonymize. Type :- Pandas DataFrame
          QI_attr :- Column names of Quasi Identifiers. Type :- list
          Sensitive_attr :- Column names of Sensitive Columns
          verbose : Log details(1) or not(0).Default value is 1. Type : Boolean
          max_iter :- The maximum iteration number of joining clusters. Default value is 5 .Type :- Integer
          max_cluster_distance :- The maximum value in cluster distance. Default value is 20. Type :- Interger
        '''
        self.nan_replacement_int = 0
        self.nan_replacement_str = ''
        InputValidator.validate_input(df, QI_attr, Sensitive_attr, cat_indecies, verbose, max_iter,
                                      anonimize_ratio, max_cluster_distance, self.nan_replacement_int, self.nan_replacement_str)
        self.df = df
        self.df_copy = df.copy()
        self.df_second_copy = df.copy()
        self.QI_attr = QI_attr
        self.Sensitive_attr = Sensitive_attr
        self.n_clusters = 0
        self.verbose = verbose
        self.centroids = None
        self.less_centroids = None
        self.k_centroids = None
        self.k = 0
        self.max_iter = max_iter
        self.anonimize_ratio = anonimize_ratio
        self.max_cluster_distance = max_cluster_distance
        self.cluster_distances = None
        self.factor = 20

    def anonymize(self, k=10, mode='', center_type='fbcg', return_mode='Not_equal', iter=1):
        '''
        This method is used to anonymize the Dataset
        Parameters :- 
            k = Number of rows that cannot be distinguished from each other
            mode = if this is 'kmode', clustering will happen using KMODE clustering. Else it will happen in using Pandas Dataframe functions.
            center_type = Defines the method to choose cluster centers.
                  If method is not equal to kmode. Three values are possible.
                      1. 'fcbg' = This method return cluster centroids weight on the probability of row's column values appear in dataframe. Default Value.
                      2. 'rsc'  = This method will choose centroids weight according to the column that has most number of unique values.
                      3. 'random = Return cluster centroids in randomly.
            return_mode = If this value equal to 'equal' ; K anonymization will done with equal member clusters. Default value is 'Not_Equal'

        return :- 
            Anonymize dataset. Type :- Pandas DataFrame

        '''
        if (k <= 0):
            k = 10
        gv.k_global(k)
        self.k = int(k)
        self.n_clusters = len(self.df)//k
        self.c_centroids = np.zeros(len(self.df))
        if(self.verbose):
            print('K :' + str(k))
            print('Mode :' + str(mode))

        data = self.df.copy()
        unique_rows = data[gv.GV['QI']].drop_duplicates()
        mode = mode.lower()
        center_type = center_type.lower()
        return_mode = return_mode.lower()

        if(mode != 'komde'):
            if(self.verbose):
                print("Initializing centroidds")
            if(center_type == 'rsc'):
                self.centroids = self._random_sample_centroids(unique_rows)
            elif(center_type == 'random'):
                self.centroids = self._select_centroids_using_weighted_column(
                    unique_rows)
            else:
                self.centroids = self._find_best_cluster_gens(unique_rows)
            self.centroids.reset_index(drop=True, inplace=True)

        while(iter != 0):
            global row_number
            row_number = 0
            iter = iter-1
            if(mode != 'kmode'):
                if(self.verbose):
                    print('Clustering...')
                self.df['cluster_number'] = self.df.apply(
                    lambda row: self._clustering1(row), axis=1)
                if(return_mode == "equal"):
                    self._adjust_big_clusters1()
                    self._adjust_small_clusters1()
            else:
                if(center_type not in ['hung,cao']):
                    center_type = 'random'
                self.df = self._komode_clustering(
                    catergorical_indexes=gv.GV['CAT_INDEXES'], type_=center_type, n_init_=self.max_iter, verbose_=self.verbose)

            self._make_anonymize()
            self.anon_k_clusters()
            self.df[gv.GV['QI']] = self.df_second_copy[gv.GV['QI']]
            self.file_write()
            return self.df[gv.GV['QI']+gv.GV['SA']].applymap(str)

    def data_loss(self):
        """
        return complete data_loss
        input is the anonymized dataframe
        ouput is number between 0 and 1
        """
        return Dataloss.complete_data_loss(self.df, self.factor)

    # def _validate_input(self,df,QI_attr,Sensitive_attr,verbose,max_iter,anonimize_ratio,max_cluster_distance):
    #     """
    #     check input validity of the user inputs
    #     """
    #     InputValidator.validate_input(selfdf,QI_attr,Sensitive_attr,verbose,max_iter,anonimize_ratio,max_cluster_distance)
    #     return True

    def _level_cluster(self, cluster_num):
        """
        cluster_num is the number assigned to 
        """
        print(cluster_num)
        Clustering.level_cluster(self.df, cluster_num)

    def _adjust_big_clusters1(self):
        """
        Comment
        """
        Clustering.adjust_big_clusters1(self.df)

    def _adjust_small_clusters1(self):
        """
        Comment
        """
        num_of_iter = 0
        while(not(self.df['cluster_number'] != -1).all()):
            num_of_iter += 1
            df = self.df.loc[self.df['cluster_number'] == -1]
            best_clusters = df.apply(lambda row: self._clustering2(row), axis=1)
            self.df.at[best_clusters.index,
                       'cluster_number'] = best_clusters.values
            self._adjust_big_clusters1()
            if(num_of_iter >= self.max_iter):
                self.df = self.df.loc[self.df['cluster_number'] != -1]
                break

    def _clustering1(self, row):
        """
        Comment
        """
        global row_number
        row_number += 1
        best_cluster = Clustering.find_best_cluster(self.df, row, self.centroids)
        return best_cluster

    def _clustering2(self, row):
        """
        Comment
        """
        temp = self.df['cluster_number'].value_counts()
        small_clusters = temp.loc[temp < gv.k].index
        if (-1 in small_clusters):
            small_clusters.drop(-1)
        best_cluster = Clustering.find_best_cluster(
            self.df, row, self.centroids.iloc[small_clusters])
        return best_cluster

    def _komode_clustering(self, catergorical_indexes=[], type_='random', n_init_=3, verbose_=0):
        '''
        This method is used to clustering based on KMODE clustering.

        dataset : The pandas dataframe want to cluster. Type :- Pandas DataFrame
        catergorical_indexes = index of columns with catergorical type data. Type List 
        type_ = Method to clustering - Hung or Cao or random. Type String
        n_clusters_ = number of clusters. Type :- Integer 
        n_init_ = number of iterations to compare. Type :- Integer 
        verbose : Log details(1) or not(0). Type : Boolean
        '''
        km = KModes(self.n_clusters, init=type_,
                    n_init=n_init_, verbose=verbose_)
        y = km.fit_predict(self.df[gv.GV['QI']],
                           categorical=catergorical_indexes)
        columns = pd.DataFrame(km.cluster_centroids_, columns=gv.GV['QI'])

        columns[gv.GV['NUM_COL']] = columns[gv.GV['NUM_COL']
                                            ].applymap(lambda x: float(x))
        self.df[gv.GV['NUM_COL']] = self.df[gv.GV['NUM_COL']
                                            ].applymap(lambda x: float(x))
        columns[gv.GV['CAT_COL']] = columns[gv.GV['CAT_COL']
                                            ].applymap(lambda x: str(x))
        self.df[gv.GV['CAT_COL']] = self.df[gv.GV['CAT_COL']
                                            ].applymap(lambda x: str(x))
        self.df['cluster_number'] = list(km.labels_)
        non_zero_member_cluster_indices = self.df.groupby('cluster_number').filter(
            lambda grp: len(grp) != 0)['cluster_number'].unique()
        columns = columns.loc[non_zero_member_cluster_indices]
        columns = columns.reset_index()
        index_series = pd.Series(
            columns.index, index=non_zero_member_cluster_indices)
        self.c_centroids = columns
        self.df['cluster_number'] = self.df.apply(
            lambda row: index_series.loc[row['cluster_number']], axis=1)
        return self.df

    def _find_best_cluster_gens(self, dataframe):
        '''
        This method return cluster centroids weight on the probability of row's column values appear in dataframe.
        dataframe :- Unique rows in the dataframe. Type :- Pandas Dataframe
        '''

        return ClusterInit.find_best_cluster_gens(self.n_clusters, dataframe)

    def _select_centroids_using_weighted_column(self, dataframe):
        '''
        This method will choose centroids weight according to the column that has most number of unique values.
        dataframe :- Unique rows in the dataframe. Type :- Pandas Dataframe
        '''
        return ClusterInit.select_centroids_using_weighted_column(self.n_clusters, dataframe)

    def _random_sample_centroids(self, unique_rows):
        '''
        Return cluster centroids in randomly.
        dataframe :- Unique rows in the dataframe. Type :- Pandas Dataframe
        '''
        return ClusterInit.random_sample_centroids(self.n_clusters, unique_rows)

    def _make_anonymize(self, method='dataloss'):
        '''
        This method is used to generalize the dataframe after clustering
        method :- If this is dataloss, Distance of two clusters is measured by the dataloss of joining, else Distance between cluster centroids. Type :- String
        '''
        result = self._mark_clusters(method='dataloss')
        if(result == 1):
            return 1
        else:
            self.mark_less_clusters_to_kclusters()

    def _mark_clusters(self, method='dataloss'):
        '''
        This method is used join less member clusters to nearest cluster.
        method :- If this is dataloss, Distance of two clusters is measured by the dataloss of joining, else Distance between cluster centroids. Type :- String
        '''
        self.df_second_copy = self.df.copy()
        if(method == 'dataloss'):
            self.cluster_distances = self._cluster_data_loss()
        else:
            self.cluster_distances = self.less_centroids.apply(
                lambda row: self.get_distance_centers(row), axis=1)
        iteration_num = 0
        while(True):
            less_groups = self.df_second_copy.groupby('cluster_number').filter(
                lambda x: len(x) < self.k).groupby('cluster_number')
            if(less_groups.ngroups == 0):
                return 1
            elif(iteration_num >= self.max_iter):
                return 0
            else:
                # if(_DEBUG):                                                                                       ####################################################
                # self.mark_less_n_kcentroids()
                self.mark_less_clusters_to_close_clusters(self)
                iteration_num += 1

    def mark_less_n_kcentroids(self, dataframe='second'):
        '''
        This method is used to mark cluster centroids which has less than k number of members.
        dataframe :- The dataframe is used to couunt the cluster members. Type Pandas Dataframe
        '''
        if(dataframe == 'second'):
            dataframe = self.df_second_copy
        else:
            dataframe = self.df
        temp = dataframe['cluster_number'].value_counts()
        self.less_centroids = self.centroids.loc[temp.loc[temp <
                                                          self.k*self.anonimize_ratio].index]
        self.k_centroids = self.centroids.loc[temp.loc[temp >=
                                                       self.k*self.anonimize_ratio].index]

    def get_distance_centers(self, cluster_):
        '''
        This method is used to find the distances between cluster generalization values.
        cluster_ :- The cluster generalization value that need to find the distance. Type :- Pandas Series
        '''
        categorical_col = self.centroids[gv.GV['CAT_COL']]
        numerical_col = self.centroids[gv.GV['NUM_COL']]
        ranges = numerical_col.max() - numerical_col.min()
        return np.sum(Calculator.cal_num_col_dist(cluster_[gv.GV['NUM_COL']], numerical_col, ranges, 20), axis=1) + Calculator.cal_cat_col_dist3(cluster_, categorical_col)

    def _cluster_data_loss(self, apply_for='less_clusters', initialize=True):
        '''
        This method is used to find the dataloss of joining two clusters.
        This function return the dataloss among each and every clusters
        apply_for :- If this is 'less_clusters', Dataloss will find only for clusters that have less members than k. Type :- String
        initialize :- This parameter define Is it essential to initialize k less clusters or not. Type :- Boolean
        '''
        categorical_dataloss = np.vectorize(
            Calculator.categorical_dataloss, excluded="cluster_list")
        if(initialize):
            self.mark_less_n_kcentroids()
        self.less_centroids.sort_index(inplace=True)
        center_groups = self.df.groupby('cluster_number')
        center_num = center_groups[gv.GV['NUM_COL']]
        center_cat = center_groups[gv.GV['CAT_COL']]
        groups = center_groups.apply(lambda x: np.unique(
            np.concatenate(x[gv.GV['CAT_COL']].values).astype(str)))
        groups = np.array(groups)
        groups = groups.reshape((groups.shape[0], 1))
        ranges = center_num.max() - center_num.min() + gv.GV['RANGE_FIX']
        if(apply_for == 'less_clusters'):
            less_groups = self.df.groupby('cluster_number').filter(
                lambda x: len(x) < self.k).groupby('cluster_number')
            if(less_groups.ngroups == 0):
                return None
            less_lists = less_groups.apply(lambda x: np.unique(
                np.concatenate(x[gv.GV['CAT_COL']].values).astype(str)))
            less_lists = np.array(less_lists)
            less_lists = less_lists.reshape((less_lists.shape[0], 1))
            cat_distances = np.apply_along_axis(
                categorical_dataloss, 1, less_lists, groups)
            num_distance = less_groups.apply(lambda row: Calculator.numerical_dataloss(
                row[gv.GV['NUM_COL']], center_num, ranges))
            cat_frame_indices = self.less_centroids.index
        else:
            cat_distances = np.apply_along_axis(
                categorical_dataloss, 1, groups, groups)
            num_distance = center_groups.apply(lambda row: Calculator.numerical_dataloss(
                row[gv.GV['NUM_COL']], center_num, ranges))
            cat_frame_indices = self.centroids.index
        shape = cat_distances.shape
        cat_distances = cat_distances.reshape(shape[0], shape[1]*shape[2])
        cat_frame = pd.DataFrame(cat_distances, index=cat_frame_indices)
        return cat_frame.add(num_distance, fill_value=gv.GV['QI_LEN']*self.max_cluster_distance)

    def mark_less_clusters_to_close_clusters(self, method='dataloss'):
        '''
        This method is used join less member clusters to nearest cluster.
        method :- If this is dataloss, Distance of two clusters is measured by the dataloss of joining, else Distance between cluster centroids. Type :- String
        '''
        self.mark_less_n_kcentroids()
        try:
            n_close_centroids = np.argsort(
                self.cluster_distances, axis=1).iloc[:, 1]
        except IndexError:
            return 1
        less_groups = self.df.groupby(
            'cluster_number').filter(lambda x: len(x) < self.k)
        groups = less_groups.groupby('cluster_number')
        less_cluster_indices = less_groups.index

        self.df_second_copy.at[less_cluster_indices, 'cluster_number'] = groups.apply(
            lambda grp: Kmodehelpers.edit_cluster(grp, n_close_centroids))

    def mark_less_clusters_to_kclusters(self, method='dataloss'):
        '''
        This method is used join less member clusters to nearest k or more member cluster.
        method :- If this is dataloss, Distance of two clusters is measured by the dataloss of joining, else Distance between cluster centroids. Type :- String
        '''
        self.mark_less_n_kcentroids()
        k_indices = self.df_second_copy.groupby('cluster_number').filter(
            lambda x: len(x) >= self.k)['cluster_number'].unique()
        cluster_distances = self.cluster_distances
        cluster_distances = cluster_distances[k_indices]

        try:
            if(k_indices.size == 1):
                n_close_centroids = np.argsort(
                    cluster_distances, axis=1).iloc[:, 0]
            else:
                n_close_centroids = np.argsort(
                    cluster_distances, axis=1).iloc[:, 1]
            cols = cluster_distances.columns
            n_close_centroids = n_close_centroids.apply(lambda row: cols[row])

        except IndexError:
            return 1
        less_groups = self.df_second_copy.groupby(
            'cluster_number').filter(lambda x: len(x) < self.k)
        groups = less_groups.groupby('cluster_number')
        less_cluster_indices = less_groups.index
        self.df_second_copy.at[less_cluster_indices, 'cluster_number'] = groups.apply(
            lambda grp: Kmodehelpers.edit_cluster(grp, n_close_centroids))

    def file_write(self, file_name='output.csv', sep_=',', encoding_='utf-8'):
        self.df[gv.GV['QI']+gv.GV['SA']
                ].to_csv(file_name, sep=sep_, encoding=encoding_)

    def anon_k_clusters(self):
        '''
        This method is used to generalize clusters.
        '''
        groups = self.df_second_copy.groupby('cluster_number')
        num_vals = groups[gv.GV['NUM_COL']].apply(
            Kmodehelpers.numeric_range).applymap(str)
        cat_vals = groups[gv.GV['CAT_COL']].apply(
            lambda row: row.apply(Kmodehelpers.catergorical_range))
        if(gv.GV['NUM_COL'] != []):
            anom_vals = num_vals.join(cat_vals)[gv.GV['QI']]
        else:
            anom_vals = cat_vals.join(num_vals)[gv.GV['QI']]
        self.df_second_copy[gv.GV['QI']] = self.df_second_copy.apply(
            lambda row: anom_vals.loc[row['cluster_number']], axis=1)

    def set_nan_replacement_int(self, replacement):
        self.nan_replacement_int = replacement

    def set_nan_replacement_str(self, replacement):
        self.nan_replacement_str = replacement


class LDiversityAnonymizer():
    def __init__(self, df, quasi_identifiers, sensitive_attributes, write_to_file=False, verbose=1):
        InputValidator.L_Diverse_Validate(
            df, quasi_identifiers, sensitive_attributes)
        self.df = df[quasi_identifiers + sensitive_attributes]
        self.sensitive_attributes = sensitive_attributes
        self.quasi_identifiers = quasi_identifiers
        self.verbose = verbose

########### Methods to write pandas dataframe to a file  #########################

    def file_write(self, file_name='output.csv', sep_=',', encoding_='utf-8'):
        self.df[self.quasi_identifiers +
                self.sensitive_attributes].to_csv(file_name, sep=sep_, encoding=encoding_)

########## Method to perform L Diversity #########################################

    def make_anonymize(self):
        l_diverse_rows = self.df.groupby(self.quasi_identifiers).filter(
            lambda group: self.count_sensitive(group))
        self.df = l_diverse_rows[self.quasi_identifiers +
                                 self.sensitive_attributes]

    def count_sensitive(self, grp):
        accept = True
        for column in self.sensitive_attributes:
            accept = accept and len(grp[column].unique()) >= self.l
        return accept

######### Public method  ########################################################

    def anonymize(self, l=2):
        if(l < 2):
            l = 2
        self.l = int(l)
        self.make_anonymize()
        return self.df


class TClosenessAnonymizer():

    def __init__(self, df, quasi_identifiers, sensitive_attributes, write_to_file=False, verbose=1):
        InputValidator.L_Diverse_Validate(
            df, quasi_identifiers, sensitive_attributes)
        self.df = df[quasi_identifiers + sensitive_attributes]
        self.sensitive_attributes = sensitive_attributes
        self.quasi_identifiers = quasi_identifiers
        self.verbose = verbose
        self.thresholds = None

########### Methods to write pandas dataframe to a file  #########################

    def file_write(self, file_name='output.csv', sep_=',', encoding_='utf-8'):
        self.df[self.quasi_identifiers +
                self.sensitive_attributes].to_csv(file_name, sep=sep_, encoding=encoding_)

########## Method to perform T Closeness #########################################

    def make_anonymize(self):
        self.define_thresholds()
        t_closeness_rows = self.df.groupby(self.quasi_identifiers).filter(
            lambda group: self.check_thresholds(group))
        self.df = t_closeness_rows[self.quasi_identifiers +
                                   self.sensitive_attributes]

    def define_thresholds(self):
        thresholds = {column: self.df[column].value_counts(
        )/len(self.df) for column in self.sensitive_attributes}
        self.thresholds = thresholds

    def check_thresholds(self, grp):
        length_cluster = len(grp)
        accept = True
        for column in self.sensitive_attributes:
            grp_thresholds = grp[column].value_counts()/length_cluster
        for element in grp_thresholds.keys():
            accept = accept and (
                grp_thresholds[element]) + self.t >= self.thresholds[column][element]
        return accept

######### Public method  ########################################################

    def anonymize(self, t=0.2):
        if(t >= 1):
            self.t = 0.2
        else:
            self.t = t
        self.make_anonymize()

        return self.df



from .. import gv
from .distance_calculation import Calculator as cl

class Clustering:

    @staticmethod
    def level_cluster(df,cluster_num):
        cluster_len = df.loc[df['cluster_number']==cluster_num].shape[0]
        additional_entries = df.loc[df['cluster_number']==cluster_num].nlargest(cluster_len-gv.k,'cluster_distance')
        df.at[additional_entries.index,'cluster_number'] = -1
        df.at[additional_entries.index,'cluster_distance'] = -1 

    @staticmethod
    def adjust_big_clusters1(df):
        temp = df['cluster_number'].value_counts()
        big_clusters = temp.loc[temp > gv.k].index
        for index in big_clusters:
            Clustering.level_cluster(df,index)


    @staticmethod
    def find_best_cluster(df,row,clusters,factor= 20):
        """
        calcuate the distance from a data point to all the cluster centroids
        and find the best cluster
        
        row - datapoint as pandas series from the dataframe, 
        clusters - cluster dataframe with all the cluster centroids
        factor - is to change priority between numerical data points and categorical data points
        when you give a higher value for the factor numerical data points get higher priority 
        resulting lower ranges for a cluster. but this cause higher data loss for categorical data

        returns the index/cluster_number that best fit to the given row.
        """

        numerical_col = clusters[gv.GV['NUM_COL']]
        categorical_col = clusters[gv.GV['CAT_COL']]
        ranges = numerical_col.max() - numerical_col.min()
        numerical_distance = cl.cal_num_col_dist(row[gv.GV['NUM_COL']],numerical_col,ranges,factor)
        categorical_distance = cl.cal_cat_col_dist3(row,categorical_col)
        df.at[row.name,'cluster_distance'] = (numerical_distance.sum(axis=1)+categorical_distance).min()
        return (numerical_distance.sum(axis=1)+categorical_distance).idxmin()

    



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="privacy_preserver",  # Replace with your own username
    version="1.0.1",
    author="telesoho",
    author_email="telesoho@gmail.com",
    description="Anonymizing Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/telesoho/privacy-preserver",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'pandas>=1.3.2',
        'numpy>=1.21.2',
        'kmodes'
    ]
)


# from pyspark.sql.types import *
# from pyspark.sql.functions import PandasUDFType, lit, pandas_udf
from .clustering_anonymizer import Kanonymizer, LDiversityAnonymizer, TClosenessAnonymizer

class Preserver:
    @staticmethod
    def k_anonymize(df, schema, QI, SA, CI, k, mode='', center_type='fbcg', return_mode='Not_equal', iter=1):

        @pandas_udf(schema, PandasUDFType.GROUPED_MAP)
        def anonymizer(pdf):
            anonymizer = Kanonymizer(pdf, QI, SA, CI)
            a_df = anonymizer.anonymize(
                k=k, mode=mode, center_type=center_type, return_mode=return_mode, iter=iter)
            return a_df

        return df.groupby().apply(anonymizer)

    @staticmethod
    def l_diverse(df, schema, quasi_identifiers, sensitive_attributes, write_to_file=False, l=2):

        @pandas_udf(schema, PandasUDFType.GROUPED_MAP)
        def anonymizer(pdf):

            anonymizer = LDiversityAnonymizer(
                pdf, quasi_identifiers, sensitive_attributes, write_to_file)
            a_df = anonymizer.anonymize(l=l)
            return a_df

        return df.groupby().apply(anonymizer)

    @staticmethod
    def t_closer(df, schema, quasi_identifiers, sensitive_attributes, t=0.3, write_to_file=False, verbose=1):

        @pandas_udf(schema, PandasUDFType.GROUPED_MAP)
        def anonymizer(pdf):
            anonymizer = TClosenessAnonymizer(
                pdf, quasi_identifiers, sensitive_attributes, write_to_file)
            a_df = anonymizer.anonymize(t=t)
            return a_df

        return df.groupby().apply(anonymizer)

    @staticmethod
    def test(df,QI, SA, CI, k,mode='', center_type='fbcg', return_mode='Not_equal', iter=1):
        anonymizer = Kanonymizer(df,QI,SA,CI)
        df = anonymizer.anonymize(k=k, mode=mode, center_type=center_type, return_mode=return_mode, iter=iter)
        return df